package mw;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
//import pnp.persistent.*;

/**
 * <pre>
 * This class  models a DELTA  character file as defined in the 
 * Definition of the Delta Format
 * 14 March, 1995
 * Dallwitz & Paine 
 *
 * We have substituted the name ItemCharacter for Character,
 * for convenience, since Java already has a Character class.
 *
 * </pre>
 * @author Noel Cross
 * @author Michael Bartley
 */

public class ItemCharacter extends BasicItemCharacter implements Persistent {
  Vector states;

  private static Hashtable cache = new Hashtable();

  //boolean isMandatory

  public ItemCharacter() {
    super();
    states = new Vector();
  }

  public void setId(Object o){
    if(o instanceof Integer){
      Integer id = (Integer)o;
      this.id = id;
      if(id != null){
	cache.put(this.id, this);
      }
    }
  }
  public void setFeature(String s){
    super.setFeature(s);
  }
  public void setStates(Vector states){this.states = states;}
  public void addState(State state){
    states.addElement(state);
  }

  public void removeState(State state){states.removeElement(state);}

  public Vector getStates(){return states;}
  public MultiState getMultiState(Integer id){
    for(int i = 0; i < states.size(); i++){
      MultiState ms = (MultiState)states.elementAt(i);
      if(ms.getId().equals(id)){
	return ms;
      }
    }
    // if no matching id is found then:
    return null;
  }
  public void print() {
    System.out.println(toString());
  }  
  public String toString() {
    StringBuffer sb = new StringBuffer();
    if(id != null) {
      sb.append("id :" + id);
    }
    if(type != null) {
      sb.append("type :" + type);
    }
    if(feature != null) {
      sb.append("feature :" + feature);
    }
    if(states != null) {
      sb.append("There are " + states.size() + " states:");   
      for(int i = 0; i < states.size(); i++){
	State s = (State)states.elementAt(i);
	sb.append(s.toString());
      }
    }
    return new String(sb);
  }

  // Persistence stuff
  private static PersistentPeer thePeer;
  public static void setPersistentPeer(PersistentPeer peer) {thePeer = peer;}
  public PersistentPeer getPersistentPeer() {return thePeer;}

  public static ItemCharacter restore(Integer id) {
    ItemCharacter ic = (ItemCharacter)cache.get(id);
    if(ic == null){
      ic = new ItemCharacter();
      ic.setId(id);
      ic.restore();
    }
    //try again.
    ic = (ItemCharacter)cache.get(id);
    return ic;
  }
  public static void restoreAll(){
    System.out.println("Start restoreAll");
    ItemCharacter ic = new ItemCharacter();
    ic.setId(null);
    ic.restore();
    System.out.println("End restoreAll");
  }
  public static Enumeration elements(){
    return cache.elements();
  }

  public void restore() {
    try{
      thePeer.restore(this);
    }catch(NothingRestoredException e){}
  }
  public void update() {thePeer.update(this);}
  public void insert() {thePeer.insert(this);}
  public void delete() {thePeer.delete(this);}

}
